/*****************************************************************************

    TRAVIS - Trajectory Analyzer and Visualizer

    http://www.travis-analyzer.de/

    Copyright (c) 2009-2018 Martin Brehm
                  2012-2018 Martin Thomas
                  2016-2018 Sascha Gehrke

    This file was written by Sascha Gehrke.

    ---------------------------------------------------------------------------

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*****************************************************************************/

// This must always be the first include directive
#include "config.h"

#include "reflux.h"
#include "luzar.h"
#include "math.h"

#include "globalvar.h"
#include "maintools.h"

static CxObArray g_ReFluxObserv;

bool gatherReFlux()
{
	CReFluxObservation *o;

	try { o = new CReFluxObservation(); } catch(...) { o = NULL; }
	if (o == NULL) NewException((double)sizeof(CReFluxObservation),__FILE__,__LINE__,__PRETTY_FUNCTION__);

	g_ReFluxObserv.Add(o);

	return true;
}

bool initializeReFlux() 
{
	int i;

	for(i = 0; i < g_ReFluxObserv.GetSize(); i++) 
	{
		((CReFluxObservation *)g_ReFluxObserv[i])->initialize();
	}
	return true;
}

bool processReFlux(CTimeStep* ts) 
{
	int i;
	for(i = 0; i < g_ReFluxObserv.GetSize(); i++) 
	{
		((CReFluxObservation *)g_ReFluxObserv[i])->process(ts);
	}
	return true;
}

bool finalizeReFlux() 
{
	int i;
	for(i = 0; i < g_ReFluxObserv.GetSize(); i++) 
	{
		((CReFluxObservation *)g_ReFluxObserv[i])->finalize();
	}
	return true;
}

CReFluxObservation::CReFluxObservation()
{
	double d_temp;
	int i,j,z;
	CxString buf, buf2;
	CReFluxCond* temp;

	if (g_oaMolecules.GetSize() > 1)
	{
		buf.sprintf("\n    Which of the molecules should be the reference molecule (");
		for (z=0;z<g_oaMolecules.GetSize();z++)
		{
			buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
			buf.strcat(buf2);
			if (z < g_oaMolecules.GetSize()-1)
				buf.strcat(", ");
		}
		buf.strcat(")? ");
		m_iShowMol = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
		mprintf(WHITE,"\n    %s is the reference molecule.\n\n",((CMolecule*)g_oaMolecules[m_iShowMol])->m_sName);

		buf.sprintf("\n    Which of the molecules should be the observed molecule (");
		for (z=0;z<g_oaMolecules.GetSize();z++)
		{
			buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
			buf.strcat(buf2);
			if (z < g_oaMolecules.GetSize()-1)
				buf.strcat(", ");
		}
		buf.strcat(")? ");
		m_iShowMol2 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
		mprintf(WHITE,"\n    %s is the observed molecule.\n\n",((CMolecule*)g_oaMolecules[m_iShowMol2])->m_sName);

	} else 
	{
		m_iShowMol = 0; 	
		m_iShowMol2 = 0;		
	}
	for ( i=0; i < ((CMolecule*)g_oaMolecules[m_iShowMol])->m_laSingleMolIndex.GetSize(); i++)
	{
		for ( j=0; j < ((CMolecule*)g_oaMolecules[m_iShowMol])->m_laSingleMolIndex.GetSize(); j++)
		{
			try { temp = new CReFluxCond(); } catch(...) { temp = NULL; }
			if (temp == NULL) NewException((double)sizeof(CReFluxCond),__FILE__,__LINE__,__PRETTY_FUNCTION__);
			m_oaCond.Add(temp);
		}
	}
	mprintf("\n    Analyzing %d pairs of molecules!\n", m_oaCond.GetSize());

	mprintf("    Which atom from %s defines the neighboring condition (e.g. \"C1,O3,#2\" )?",((CMolecule*)g_oaMolecules[m_iShowMol])->m_sName);
	inpprintf("! Which atom from %s defines the neighboring condition (e.g. \"C1,O3,#2\" )?\n",((CMolecule*)g_oaMolecules[m_iShowMol])->m_sName);
	myget(&buf);
	while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
	{
		mprintf("    Please enter single atom!\n");
		myget(&buf);
	}

	mprintf("    Which atom from %s defines the neighboring condition (e.g. \"C1,O3,#2\" )?",((CMolecule*)g_oaMolecules[m_iShowMol2])->m_sName);
	inpprintf("! Which atom from %s defines the neighboring condition (e.g. \"C1,O3,#2\" )?\n",((CMolecule*)g_oaMolecules[m_iShowMol2])->m_sName);
	myget(&buf);
	while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
	{
		mprintf("    Please enter atom!\n");
		myget(&buf);
	}
	d_temp = AskFloat("    Enter maximum distance for neighbor criterion [300] ",300.0);
	for ( i=0; i < m_oaCond.GetSize(); i++ )
		((CReFluxCond*)m_oaCond[i])->m_dNeighDist = d_temp;

	if ( !AskYesNo("    Define a distance criterion for the aggregation? [yes]", true) )
		((CReFluxCond*)m_oaCond[0])->m_dAggrDist = -2.0;
	else
	{
		mprintf("    Use which atom(s) from %s to distance condition (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol])->m_sName);
		inpprintf("! Use which atom(s) from %s to distance condition (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol])->m_sName);
		myget(&buf);
		while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
		{
			mprintf("    Please enter atom(s)!\n");
			myget(&buf);
		}
		mprintf("    Use which atom(s) from %s to distance condition (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol2])->m_sName);
		inpprintf("! Use which atom(s) from %s to distance condition (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol2])->m_sName);
		myget(&buf);
		while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
		{
			mprintf("    Please enter atom(s)!\n");
			myget(&buf);
		}
		d_temp = AskFloat("    Enter maximum distance for aggregation criterion [300] ",300.0);
		for ( i=0; i < m_oaCond.GetSize(); i++ )
			((CReFluxCond*)m_oaCond[i])->m_dAggrDist = d_temp;
	}
	if ( !AskYesNo("    Define an angle criterion for the aggregation? [yes]", true) )
		((CReFluxCond*)m_oaCond[0])->m_dAggrMaxAngle = -2.0;
	else
	{
		m_bRefVecPlane = !AskYesNo("    Define first vector by two points (y) or perpendicular to a plane (n) ? [yes]", true);

		if (g_oaMolecules.GetSize() > 1)
		{
			buf.sprintf("\n    Take the base of the vector from wich molecule (");
			for (z=0;z<g_oaMolecules.GetSize();z++)
			{
				buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
				buf.strcat(buf2);
				if (z < g_oaMolecules.GetSize()-1)
					buf.strcat(", ");
			}
			buf.strcat(")? ");
			m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
		}
		mprintf("    Use which atom(s) from %s as vector base (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
		inpprintf("! Use which atom(s) from %s as vector base (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
		myget(&buf);
		while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
		{
			mprintf("    Please enter atom(s)!\n");
			myget(&buf);
		}
		
		if ( m_bRefVecPlane )
		{
			if (g_oaMolecules.GetSize() > 1)
			{
				buf.sprintf("\n    Take the tipp of first vector from wich molecule (");
				for (z=0;z<g_oaMolecules.GetSize();z++)
				{
					buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
					buf.strcat(buf2);
					if (z < g_oaMolecules.GetSize()-1)
						buf.strcat(", ");
				}
				buf.strcat(")? ");
				m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
			}
			mprintf("    Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			inpprintf("! Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			myget(&buf);
			while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
			{
				mprintf("    Please enter atom(s)!\n");
				myget(&buf);
			}
			if (g_oaMolecules.GetSize() > 1)
			{
				buf.sprintf("\n    Take the tipp of second vector from wich molecule (");
				for (z=0;z<g_oaMolecules.GetSize();z++)
				{
					buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
					buf.strcat(buf2);
					if (z < g_oaMolecules.GetSize()-1)
						buf.strcat(", ");
				}
				buf.strcat(")? ");
				m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
			}
			mprintf("    Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			inpprintf("! Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			myget(&buf);
			while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
			{
				mprintf("    Please enter atom(s)!\n");
				myget(&buf);
			}
		}
		else
		{
			if (g_oaMolecules.GetSize() > 1)
			{
				buf.sprintf("\n    Take the tipp of the vector from wich molecule (");
				for (z=0;z<g_oaMolecules.GetSize();z++)
				{
					buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
					buf.strcat(buf2);
					if (z < g_oaMolecules.GetSize()-1)
						buf.strcat(", ");
				}
				buf.strcat(")? ");
				m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
			}
			mprintf("    Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			inpprintf("! Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			myget(&buf);
			while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
			{
				mprintf("    Please enter atom(s)!\n");
				myget(&buf);
			}
		}
		m_bObsVecPlane =  !AskYesNo("    Define second vector by two points (y) or perpendicular to a plane (n) ? [yes]", true);

		if (g_oaMolecules.GetSize() > 1)
		{
			buf.sprintf("\n    Take the base of the vector from wich molecule (");
			for (z=0;z<g_oaMolecules.GetSize();z++)
			{
				buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
				buf.strcat(buf2);
				if (z < g_oaMolecules.GetSize()-1)
					buf.strcat(", ");
			}
			buf.strcat(")? ");
			m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
		}
		mprintf("    Use which atom(s) from %s as vector base (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
		inpprintf("! Use which atom(s) from %s as vector base (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
		myget(&buf);
		while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
		{
			mprintf("    Please enter atom(s)!\n");
			myget(&buf);
		}
		
		if ( m_bRefVecPlane )
		{
			if (g_oaMolecules.GetSize() > 1)
			{
				buf.sprintf("\n    Take the tipp of first vector from wich molecule (");
				for (z=0;z<g_oaMolecules.GetSize();z++)
				{
					buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
					buf.strcat(buf2);
					if (z < g_oaMolecules.GetSize()-1)
						buf.strcat(", ");
				}
				buf.strcat(")? ");
				m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
			}
			mprintf("    Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			inpprintf("! Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			myget(&buf);
			while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
			{
				mprintf("    Please enter atom(s)!\n");
				myget(&buf);
			}
			if (g_oaMolecules.GetSize() > 1)
			{
				buf.sprintf("\n    Take the tipp of second vector from wich molecule (");
				for (z=0;z<g_oaMolecules.GetSize();z++)
				{
					buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
					buf.strcat(buf2);
					if (z < g_oaMolecules.GetSize()-1)
						buf.strcat(", ");
				}
				buf.strcat(")? ");
				m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
			}
			mprintf("    Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			inpprintf("! Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			myget(&buf);
			while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
			{
				mprintf("    Please enter atom(s)!\n");
				myget(&buf);
			}
		}
		else
		{
			if (g_oaMolecules.GetSize() > 1)
			{
				buf.sprintf("\n    Take the tipp of the vector from wich molecule (");
				for (z=0;z<g_oaMolecules.GetSize();z++)
				{
					buf2.sprintf("%s=%d",((CMolecule*)g_oaMolecules[z])->m_sName,z+1);
					buf.strcat(buf2);
					if (z < g_oaMolecules.GetSize()-1)
						buf.strcat(", ");
				}
				buf.strcat(")? ");
				m_iShowMol3 = AskRangeInteger_ND("%s",1,g_oaMolecules.GetSize(),(const char*)buf) - 1;
			}
			mprintf("    Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			inpprintf("! Use which atom(s) from %s as vector tipp (e.g. \"C1,O3-5,#2\" *=all)?\n",((CMolecule*)g_oaMolecules[m_iShowMol3])->m_sName);
			myget(&buf);
			while ((strlen(buf) == 0) || (!checkAtomChoice(buf)))
			{
				mprintf("    Please enter atom(s)!\n");
				myget(&buf);
			}
		}
		d_temp = cos((M_PI/180)*AskRangeFloat("    Enter minimum angle for aggregation criterion [0] ",0.0,179.0,0.0));
		for ( i=0; i < m_oaCond.GetSize(); i++ )
			((CReFluxCond*)m_oaCond[i])->m_dAggrMinAngle = d_temp;
		d_temp = cos((M_PI/180)*AskRangeFloat("    Enter maximum angle for aggregation criterion [180] ",d_temp,180.0,180.0));
		for ( i=0; i < m_oaCond.GetSize(); i++ )
			((CReFluxCond*)m_oaCond[i])->m_dAggrMaxAngle = d_temp;
	}

	for ( i=0; i < m_oaCond.GetSize(); i++ )
		((CReFluxCond*)m_oaCond[i])->Sort();

	if ( ((CReFluxCond*)m_oaCond[0])->m_iaRefAggr.GetSize() == 0 
		&& ((CReFluxCond*)m_oaCond[0])->m_iaRefVec.GetSize() == 0 )
	{
		eprintf("Error: No aggregation criterion chosen?\n");
		abort();
	}

	if ( ( ((CReFluxCond*)m_oaCond[0])->m_iaRefVec.GetSize() != ((CReFluxCond*)m_oaCond[0])->m_iaRefVec2.GetSize() 
		|| ( ((CReFluxCond*)m_oaCond[0])->m_iaRefVec.GetSize() != ((CReFluxCond*)m_oaCond[0])->m_iaRefVec3.GetSize()
			&& ((CReFluxCond*)m_oaCond[0])->m_iaRefVec3.GetSize() != 0 ) )
	|| ( ((CReFluxCond*)m_oaCond[0])->m_iaObsVec.GetSize() != ((CReFluxCond*)m_oaCond[0])->m_iaObsVec2.GetSize()
		|| ( ((CReFluxCond*)m_oaCond[0])->m_iaObsVec.GetSize() != ((CReFluxCond*)m_oaCond[0])->m_iaObsVec3.GetSize()
			&& ((CReFluxCond*)m_oaCond[0])->m_iaObsVec3.GetSize() != 0 ) ) )
	{
		eprintf("Internal Error 1: Failure in angle definition.\n");
		abort();
	}
	
	if ( ( ((CReFluxCond*)m_oaCond[0])->m_iaRefAggr.GetSize() != 0 && ((CReFluxCond*)m_oaCond[0])->m_iaRefVec.GetSize() != 0 )
		&& ( ((CReFluxCond*)m_oaCond[0])->m_iaRefAggr.GetSize() != ((CReFluxCond*)m_oaCond[0])->m_iaRefVec.GetSize() 
			|| ((CReFluxCond*)m_oaCond[0])->m_iaObsAggr.GetSize() != ((CReFluxCond*)m_oaCond[0])->m_iaObsVec.GetSize() ) )
	{
		eprintf("Internal Error 2: Failure in mixed condition.\n");
		abort();
	}
	
	g_fTimestepLength = AskFloat("    Enter the length of one trajectory time step in fs: [0.5] ",0.5f);
}

CReFluxObservation::~CReFluxObservation()
{
}

void CReFluxObservation::initialize()
{
	CxIntArray* temp;
	int i;

	for ( i=0; i < m_oaCond.GetSize(); i++ )
	{
		try { temp = new CxIntArray(); } catch(...) { temp = NULL; }
		if (temp == NULL) NewException((double)sizeof(CxIntArray),__FILE__,__LINE__,__PRETTY_FUNCTION__);
		m_oaNeigh.Add(temp);

		try { temp = new CxIntArray(); } catch(...) { temp = NULL; }
		if (temp == NULL) NewException((double)sizeof(CxIntArray),__FILE__,__LINE__,__PRETTY_FUNCTION__);
		m_oaAggr.Add(temp);
	}

	if ( g_iMaxStep != -1 )
		m_iDepth = g_iMaxStep * 3 / 4 / g_iStride;
	else
		m_iDepth = g_iTrajSteps * 3 / 4 / g_iStride;

	m_iDepth = AskUnsignedInteger("    Enter the resolution (=depth) of the ACF (in time steps): [%d] ", m_iDepth, m_iDepth);
	if ( m_iDepth > g_iTrajSteps )
	{
		mprintf("    Correlation Depth is exceeding number of total timesteps. Truncating ....\n");
		m_iDepth = g_iTrajSteps;
	}

	m_iTrunc = m_iDepth;

	while ( m_iTrunc > (int)(m_iDepth/10) )
	{
		m_iTrunc = (int)(50000/(g_fTimestepLength*g_iStride));
		if ( m_iTrunc > m_iDepth/10 )
			m_iTrunc = (int)(m_iDepth/10);	

		m_iTrunc = AskUnsignedInteger("    Enter the truncation for the fitting procedure (in time steps): [%d] ", m_iTrunc, m_iTrunc);

		if ( m_iTrunc > m_iDepth/10 )
			mprintf("    Truncation must be at least 10 times smaller than correlation depth!\n");
	}
}

void CReFluxObservation::process(CTimeStep* ts)
{
	int i;

	for ( i=0; i < m_oaCond.GetSize(); i++ )
		((CReFluxCond*)m_oaCond[i])->check(ts,((CxIntArray*)m_oaNeigh[i]),((CxIntArray*)m_oaAggr[i]));
}

void CReFluxObservation::finalize()	// reconstruct functions -> autocorrelate -> sum results -> normalize -> fit
{
	long n,i,interval;
	double AggrCor, DiffCor;
	CxString buf,buf2;
	unsigned long t0,eta;

	CxDoubleArray NeighFunction;	// Input
	CxDoubleArray AggrFunction;
	
	CxDoubleArray temp1Function;	// Temp
	CxDoubleArray temp2Function;

	CLuzarCorrelation luzar;

	for ( i=0; i < m_iDepth; i++ )
	{
		AggrDynamic.Add(0.0);
		DiffusionDynamic.Add(0.0);
	}

	NeighFunction.SetSize(g_iSteps / g_iStride);
	AggrFunction.SetSize(g_iSteps / g_iStride);

	luzar.Init(g_iSteps / g_iStride,m_iDepth); 	// Inititalize Luzar-Correlation

	mprintf("    Finishing %6i Functions ... \nFunc %6li ",m_oaNeigh.GetSize(),(long)0);
	g_iDotCounter = 0;
	t0 = (unsigned long)time(NULL);
	interval = 2;

	for ( n=0; n < m_oaNeigh.GetSize(); n++)		// Walk over all pairs
	{
		if ( n % interval == 0 )
		{
			if ( g_iDotCounter >= 50 )
			{
				if ( interval == 2 && n > 100 )
				{
					interval = 20;
				}
				else if ( interval == 20 && n > 2000 )
				{
					interval = 200;
				}
				g_iDotCounter = 0;
				if ((time(NULL) - t0) > 5)
				{
					eta = (unsigned long)(((double)time(NULL) - t0) / n * ((double)MAX((long)0,(long)m_oaNeigh.GetSize() - (long)n)));
					FormatTime(eta,&buf);
					mprintf(" ETA %s",(const char*)buf);
				}
				mprintf("\nFunc %6li ",n);
			}
			g_iDotCounter++;
			mprintf(".");
		}

		if ( ((CxIntArray*)m_oaNeigh[n])->GetSize() > 0 )
		{
			RestoreFunction(((CxIntArray*)m_oaNeigh[n]),&NeighFunction);	// Restore function of pair n as NeighFunction and AggrFunction
			RestoreFunction(((CxIntArray*)m_oaAggr[n]),&AggrFunction);	

			luzar.Correlate(&AggrFunction,&NeighFunction,&temp1Function,&temp2Function);

			for ( i=0; i < m_iDepth; i++ )		// Sum results to Output-array
			{
				AggrDynamic[i] += temp1Function[i];
				DiffusionDynamic[i] += temp2Function[i];
			}
		}
	}
	mprintf("\n\n");

	if ( AggrDynamic[0] == 0 || DiffusionDynamic[0] == 0 )
	{
		eprintf("First correlation value is zero ... something is wrong ...\n");
		abort();
	}

	AggrCor = (double)AggrDynamic[0];
	DiffCor = (double)DiffusionDynamic[0];

	for ( i=0; i < m_iDepth; i++ )			// Normalize
	{
		AggrDynamic[i] /= AggrCor;
		DiffusionDynamic[i] /= DiffCor;
		DiffusionDynamic[i] -= AggrDynamic[i];
	}

	luzar.Fit(&AggrDynamic, &DiffusionDynamic, &k_t, &a, &b, m_iTrunc);

	WriteOutput();
}


void CReFluxObservation::RestoreFunction(CxIntArray* input, CxDoubleArray* output)
{
	unsigned int n;
	bool eins = false;

	for ( n=0; n < (g_iSteps / g_iStride); n++ )		// Walk over all processed steps
	{
		if ( input->Contains((n+1)*g_iStride) )
			eins = !eins;
		if (!eins)
			output->SetAt(n,0.0);
		else
			output->SetAt(n,1.0);
	}
}

bool CReFluxObservation::includeAtom(int atom, int number, int mol, int target)	// atom: element number, number: atomsort number, mol m_iShowMol, target: in which array to include
{
	int n, m, smol, count = -1;
	CxIntArray includeME, temp;

	for ( n=0; n < ((CMolecule*)g_oaMolecules[m_iShowMol])->m_laSingleMolIndex.GetSize(); n++)
	{
		for ( m=0; m < ((CMolecule*)g_oaMolecules[m_iShowMol2])->m_laSingleMolIndex.GetSize(); m++)
		{
			count++;
			if ( mol == m_iShowMol )
				smol = n;
			else
				smol = m;

			includeME.CopyFrom((CxIntArray*)(((CSingleMolecule*)g_oaSingleMolecules[((CMolecule*)g_oaMolecules[mol])->m_laSingleMolIndex[smol]])->m_oaAtomOffset[atom]));
			if (number >= includeME.GetSize()) 
			{
				eprintf("CReFluxObservation::includeAtom(): Internal error (%d/%d).\n",number,includeME.GetSize());
				abort();
			}
			
			if 	( target == 1 )
				((CReFluxCond*)m_oaCond[count])->m_iRefNeigh = includeME[number];
			else if ( target == 2 )
				((CReFluxCond*)m_oaCond[count])->m_iObsNeigh = includeME[number];
			else if ( target == 3 )
				((CReFluxCond*)m_oaCond[count])->m_iaRefAggr.Add(includeME[number]);
			else if ( target == 4 )
				((CReFluxCond*)m_oaCond[count])->m_iaObsAggr.Add(includeME[number]);
			else if ( target == 5 )
				((CReFluxCond*)m_oaCond[count])->m_iaRefVec.Add(includeME[number]);
			else if ( target == 6 )
				((CReFluxCond*)m_oaCond[count])->m_iaRefVec2.Add(includeME[number]);
			else if ( target == 7 )
				((CReFluxCond*)m_oaCond[count])->m_iaRefVec3.Add(includeME[number]);
			else if ( target == 8 )
				((CReFluxCond*)m_oaCond[count])->m_iaObsVec.Add(includeME[number]);
			else if ( target == 9 )
				((CReFluxCond*)m_oaCond[count])->m_iaObsVec2.Add(includeME[number]);
			else if ( target == 10)
				((CReFluxCond*)m_oaCond[count])->m_iaObsVec3.Add(includeME[number]);
			else
			{
				eprintf("CReFluxObservation::includeAtom(): Internal error (target %d).\n",target);
				abort();
			}
		}
	}
	
	return true;
}


bool CReFluxObservation::checkAtomChoice(const char *s)
{
	const char *p, *q;
	char buf[32];
	const char *allowed = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz1234567890-,#*_ ";
	int inclu, la = -1, i = -1, i2 = -1, atom = -1, n, target = 0; 
	bool m = false, fin = false, single = false;
	int mol = m_iShowMol;

	if ( ((CReFluxCond*)m_oaCond[0])->m_dNeighDist == -1.0 )
		single = true;

	if 	( ((CReFluxCond*)m_oaCond[0])->m_iRefNeigh == -1 )
		target =  1;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_iObsNeigh == -1 )
		target =  2;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrDist != -2 
			&& ((CReFluxCond*)m_oaCond[0])->m_iaRefAggr.GetSize() == 0 )
		target =  3;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrDist != -2 
			&& ((CReFluxCond*)m_oaCond[0])->m_iaObsAggr.GetSize() == 0 )
		target =  4;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrMaxAngle != -2 
			&& ((CReFluxCond*)m_oaCond[0])->m_iaRefVec.GetSize() == 0 )
		target =  5;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrMaxAngle != -2 
			&& ((CReFluxCond*)m_oaCond[0])->m_iaRefVec2.GetSize() == 0 )
		target =  6;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrMaxAngle != -2 
			&& m_bRefVecPlane
			&& ((CReFluxCond*)m_oaCond[0])->m_iaRefVec3.GetSize() == 0 )
		target =  7;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrMaxAngle != -2 
			&& ((CReFluxCond*)m_oaCond[0])->m_iaObsVec.GetSize() == 0 )
		target =  8;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrMaxAngle != -2 
			&& ((CReFluxCond*)m_oaCond[0])->m_iaObsVec2.GetSize() == 0 )
		target =  9;
	else if ( ((CReFluxCond*)m_oaCond[0])->m_dAggrMaxAngle != -2 
			&& m_bObsVecPlane
			&& ((CReFluxCond*)m_oaCond[0])->m_iaObsVec3.GetSize() == 0 )
		target = 10;

	if ( target == 4 || target == 2 )
		mol = m_iShowMol2;
	else if ( target > 4 )
		mol = m_iShowMol3;
	
	p = s;

	while (*p != 0)
	{
		fin = false;
		while (*p == ' ')
			p++;
		if ( single && ( *p == ',' || *p == '*' || *p == '-' ) )
		{
			eprintf("Error: Must be single atom!\n");
			return false;
		}
		if (strchr(allowed,*p) == NULL)
		{
			eprintf("Error: Character \"%c\" not allowed.\n",*p);
			return false;
		}
		if ( *p == '*' )
		{
			for ( n=0; n < ((CMolecule*)g_oaMolecules[mol])->m_baAtomIndex.GetSize()-1; n++ )
				for (inclu = 0; inclu < ((CMolecule*)g_oaMolecules[mol])->m_waAtomCount[n]; inclu++ )
					includeAtom(n, inclu, mol, target);	 //hier
			return true;
		}
		q = p;
		if (isalpha(*q) || *q == '#')
		{
			if (m)
			{
				eprintf("Error: Only digit allowed after \"-\".\n");
				return false;
			}
			while ( isalpha(*q) || *q == '#' )
				q++;
			if (q-p >= 32)
			{
				eprintf("Internal Error (%ld >= 32).\n",q-p);
				return false;
			}
			memcpy(buf,p,q-p);
			buf[q-p] = 0;
			atom = ((CMolecule*)g_oaMolecules[mol])->FindAtomInMol(buf); 	// atom holds number of element 
			if (atom == -1)
			{
				eprintf("Error: Atom \"%s\" not in the molecule.\n",buf);
				return false;
			} 
			
		} 
		else if (isdigit(*q))
		{
			if (atom == -1)
				atom = la;
			if (atom == -1)
			{
				eprintf("Error: Number in the beginning not possible.\n");
				return false;
			}
			while (isdigit(*q)) 
				q++;
			if ((*q != '-') && (*q != ',') && (*q != ' ') && (*q != 0))
			{
				eprintf("Error: Only \",\" or \"-\" may follow after a number.\n");
				return false;
			}
			if (q-p >= 32)
			{
				eprintf("Internal Error (%ld >= 32).\n",q-p);
				return false;
			}
			memcpy(buf,p,q-p);
			buf[q-p] = 0;
			if (atoi(buf)-1 >= ((CMolecule*)g_oaMolecules[mol])->m_waAtomCount[atom])
			{
				eprintf("Error: Only %d %s atoms in the molecule (requested: %d)\n",((CMolecule*)g_oaMolecules[mol])->m_waAtomCount[atom],(const char*)((CAtom*)g_oaAtoms[((CMolecule*)g_oaMolecules[mol])->m_baAtomIndex[atom]])->m_sName,atoi(buf));
				return false;
			}
			if (m)	// Range: Include all atoms from i to i2				case: X1-3	A
			{
				i2 = atoi(buf)-1;
				if (i2 < i) 	// If range from high to low -> wrong input
				{
					eprintf("Error: Invalid atom range, %d < %d.\n",i2+1,i+1);
					return false;
				}
				else
				{
					//>>>MB170422
					//for (inclu = i; inclu < i2; inclu++)
					for (inclu = i; inclu <= i2; inclu++)
					//<<<MB170422
					{
						includeAtom(atom, inclu, mol, target);	 //hier
					}
					fin = true;
				}
			} 
			else	
			{
				i = atoi(buf)-1;
			}
		} 
		else if (*q == '-')
		{
			if (i == -1)
			{
				eprintf("Error: \"-\" without preceding number.\n");
				return false;
			} 
			m = true;
			q++;
		} 
		else if (*q == ',')   
		{
			if (atom == -1)
			{
				eprintf("Error: Comma without atom.\n");
				return false;
			}
			if ( i == -1 )	// include all atoms of kind				case: X,	C
				for (inclu = 0; inclu < ((CMolecule*)g_oaMolecules[mol])->m_waAtomCount[atom]; inclu++ )
					includeAtom(atom, inclu, mol, target);	 //hier
			else if (!m)  // include single atom							case: X1,	B
				includeAtom(atom, i, mol, target);	// hier
			la = atom;
			m = false;
			i = -1;
			i2 = -1;
			atom = -1;
			q++;
			fin = true;
		}
		p = q;
	}
	if ( !fin )
	{
		if ( i == -1 )	// case: X
		{
			for (inclu = 0; inclu < ((CMolecule*)g_oaMolecules[mol])->m_waAtomCount[atom]; inclu++ )
			{
				includeAtom(atom, inclu, mol, target);	 //hier
			}
		}
		else			// case: X1
		{
			includeAtom(atom, i, mol, target); // hier
		}
	}

	return true;
}

void CReFluxObservation::WriteOutput()
{
	CGrace gc, gc2;
	int n,m;

	mprintf("    Writing \"Correlations.agr\"...\n");
	
	gc.SetRangeX(0,m_iDepth*g_iStride*g_fTimestepLength/1000);
	gc.SetRangeY(0,1);
	gc.MakeTicks();
	gc.SetLabelX("t [ps]");
	gc.SetLabelY("f(t)");
	gc.CurrentGraph()->m_bInvertXAxis = false;
	gc.CurrentGraph()->m_bLegend = true;

	gc.AddDataset();
	gc.SetDatasetName(0, "c(t)");
	gc.AddDataset();
	gc.SetDatasetName(1, "n(t)");
	for ( n = 0; n < AggrDynamic.GetSize(); n++)
	{
		gc.AddXYTupel(0,n*g_fTimestepLength*g_iStride/1000,AggrDynamic[n]);
		gc.AddXYTupel(1,n*g_fTimestepLength*g_iStride/1000,DiffusionDynamic[n]);
	}

	gc.WriteAgr("Correlations.agr",false);
	gc.WriteCSV("Correlations.csv");

	mprintf("    Writing \"Fitted.agr\"...\n");

	CalcMinMax();

	gc2.SetRangeX(0,m_dGraceMax);
	gc2.SetRangeY(0,m_dGraceMax);
	gc2.MakeTicks();
	gc2.SetLabelX("kf c(t) - kb n(t) [1/ps]");
	gc2.SetLabelY("-dc/dt [1/ps]");
	gc2.CurrentGraph()->m_bInvertXAxis = false;
	gc2.CurrentGraph()->m_bLegend = true;
		
	gc2.AddDataset();
	gc2.SetDatasetName(0, "Data");

	for ( m=0; m < k_t.GetSize(); m++)
		gc2.AddXYTupel(0,a[0] * AggrDynamic[m] - b[0] * DiffusionDynamic[m], - k_t[m]);

	gc2.WriteCSV("Fitted.csv");
	gc2.AddDataset();
	gc2.SetDatasetName(1, "Unity");
	gc2.AddXYTupel(1,0,0);
	gc2.AddXYTupel(1,m_dGraceMax,m_dGraceMax);
	gc2.WriteAgr("Fitted.agr",false);
}

void CReFluxObservation::CalcMinMax()
{
	double temp;
	int n;	

	m_dGraceMin =  1E20;
	m_dGraceMax = -1E20;

	for ( n = m_iTrunc / 5; n < AggrDynamic.GetSize(); n++)
	{
		temp = (a[0] * AggrDynamic[n] - b[0] * DiffusionDynamic[n]);
		if ( m_dGraceMin >= temp )
			m_dGraceMin = temp;
		if ( m_dGraceMax <= temp )
			m_dGraceMax = temp;
	}

	for ( n = m_iTrunc / 5; n < k_t.GetSize(); n++)
	{
		temp = -k_t.GetAt(n);
		if ( m_dGraceMin >= temp )
			m_dGraceMin = temp;
		if ( m_dGraceMax <= temp )
			m_dGraceMax = temp;
	} 
}

CReFluxCond::CReFluxCond()
{
	m_iRefNeigh = -1;
	m_iObsNeigh = -1;
	m_dNeighDist = -1.0;
	m_dAggrDist = -1.0;
	m_dAggrMinAngle = -1.0;
	m_dAggrMaxAngle = -1.0;
}

CReFluxCond::~CReFluxCond()
{
}

void CReFluxCond::check(CTimeStep* ts, CxIntArray* na, CxIntArray* aa)
{
	int i,j;
	CxDVector3 Vec;

	if ( m_iRefNeigh == m_iObsNeigh )
		return;
	Vec = FoldVector( ts->m_vaCoords[m_iRefNeigh] - ts->m_vaCoords[m_iObsNeigh] );
	if ( Vec.GetLength() > m_dNeighDist )
	{
		if ( na->GetSize() % 2 == 1 )
			na->Add(g_iSteps);
		if ( aa->GetSize() % 2 == 1 )
			aa->Add(g_iSteps);
		return;
	}
	if ( na->GetSize() % 2 == 0 )
		na->Add(g_iSteps);

	for ( i=0; i < m_iaRefAggr.GetSize(); i++ )
	{
		for ( j=0; j < m_iaObsAggr.GetSize(); j++ )
		{
			Vec = FoldVector( ts->m_vaCoords[m_iaRefAggr[i]] - ts->m_vaCoords[m_iaObsAggr[j]] );
			if ( Vec.GetLength() < m_dAggrDist && angle(ts,i,j) )
			{
				if ( aa->GetSize() % 2 == 0 )
					aa->Add(g_iSteps);
				return;
			}	
		}
	}
	if ( m_iaRefAggr.GetSize() > 0 )
	{
		if ( aa->GetSize() % 2 == 1 )
			aa->Add(g_iSteps);
		return;
	}
	for ( i=0; i < m_iaRefVec.GetSize(); i++ )
	{
		for ( j=0; j < m_iaObsVec.GetSize(); j++ )
		{
			if ( angle(ts,i,j) )
			{
				if ( aa->GetSize() % 2 == 0 )
					aa->Add(g_iSteps);
				return;
			}
		}
	}
	if ( aa->GetSize() % 2 == 1 )
		aa->Add(g_iSteps);
}

bool CReFluxCond::angle(CTimeStep* ts, int i, int j)
{
	CxDVector3 Vec, Vec2, Vec3;
	double alpha;

	Vec  = FoldVector( ts->m_vaCoords[m_iaRefVec2[i]] - ts->m_vaCoords[m_iaRefVec[i]] );
	if ( m_iaRefVec3.GetSize() != 0 )
	{
		Vec2 = FoldVector( ts->m_vaCoords[m_iaRefVec3[i]] - ts->m_vaCoords[m_iaRefVec[i]] );
		Vec = CrossP(Vec,Vec2);
	}
	Vec2 = FoldVector( ts->m_vaCoords[m_iaObsVec2[j]] - ts->m_vaCoords[m_iaObsVec[j]] );
	if ( m_iaObsVec3.GetSize() != 0 )
	{
		Vec3 = FoldVector( ts->m_vaCoords[m_iaObsVec3[j]] - ts->m_vaCoords[m_iaObsVec[j]] );
		Vec2 = CrossP(Vec2,Vec3);
	}
	Vec.Normalize();
	Vec2.Normalize();

	alpha = DotP(Vec,Vec2);
	if ( alpha > m_dAggrMinAngle && alpha < m_dAggrMaxAngle )
		return true;
	return false;
}

void CReFluxCond::Sort()
{
	CxIntArray temp;

	if ( m_iaRefVec.GetSize() == 0 )
		return;

	if ( m_iaRefVec.GetSize() == 1 )
		while ( m_iaRefVec.GetSize() < m_iaRefAggr.GetSize() 
			|| m_iaRefVec.GetSize() < m_iaRefVec2.GetSize()
			|| m_iaRefVec.GetSize() < m_iaRefVec3.GetSize() )
			m_iaRefVec.Add(m_iaRefVec[0]);
	if ( m_iaRefVec2.GetSize() == 1 )
		while ( m_iaRefVec2.GetSize() < m_iaRefVec.GetSize() )
			m_iaRefVec2.Add(m_iaRefVec2[0]);
	if ( m_iaRefVec3.GetSize() == 1 )
		while ( m_iaRefVec3.GetSize() < m_iaRefVec.GetSize() )
			m_iaRefVec3.Add(m_iaRefVec3[0]);
	if ( m_iaRefAggr.GetSize() == 1 )
		while ( m_iaRefAggr.GetSize() < m_iaRefVec.GetSize() )
			m_iaRefAggr.Add(m_iaRefAggr[0]);

	if ( m_iaObsVec.GetSize() == 1 )
		while ( m_iaObsVec.GetSize() < m_iaObsAggr.GetSize() 
			|| m_iaObsVec.GetSize() < m_iaObsVec2.GetSize()
			|| m_iaObsVec.GetSize() < m_iaObsVec3.GetSize() )
			m_iaObsVec.Add(m_iaObsVec[0]);
	if ( m_iaObsVec2.GetSize() == 1 )
		while ( m_iaObsVec2.GetSize() < m_iaObsVec.GetSize() )
			m_iaObsVec2.Add(m_iaObsVec2[0]);
	if ( m_iaObsVec3.GetSize() == 1 )
		while ( m_iaObsVec3.GetSize() < m_iaObsVec.GetSize() )
			m_iaObsVec3.Add(m_iaObsVec3[0]);
	if ( m_iaObsAggr.GetSize() == 1 )
		while ( m_iaObsAggr.GetSize() < m_iaObsVec.GetSize() )
			m_iaObsAggr.Add(m_iaObsAggr[0]);
}
